;(function($, win) {
    'use strict';
    
    var requestFrame = (function(){
        var raf = win.requestAnimationFrame ||
            win.mozRequestAnimationFrame ||
            win.webkitRequestAnimationFrame ||
            function(fn){ return win.setTimeout(fn, 20); };
        return function(fn){ return raf(fn); };
    })();

    /**
     * Sticky Element constructor
     * @param elm {String}
     * @param par {String}
     * @param [options] {Object}
     * @constructor
     */
    var Sticky = function(elm, par, options) {
        this.element = elm;
        this.parent = par;
        this._frozen = false;
        this.options = $.extend({
            useTransition: false,
            animate: true,
            animTime: 200,
            animDelay: 300
        }, options);

        this.init();
    };

    Sticky.prototype.init = function() {
        var transition = '';
        if(this.options.useTransition){
            transition = 'top ' + this.options.animTime + 'ms ease-in-out';
            this.options.animate = false;
        }

        this.parent.css({'position':'relative'});
        this.element
            .addClass('sticky-scroll')
            .css({
                'transition': transition,
                'position':'relative'
            });

        this.update();
    };

    /**
     * This will handle any resizing of the container the sticky scroll is in and update the boundaries if necessary
     */
    Sticky.prototype.update = function() {
        this.setBoundaries(0);
        this.moveIt();
    };

    /**
     * This will decide whether to move the stickied item
     */
    Sticky.prototype.moveIt = function() {
        var scrollTop = win.document.body.scrollTop;
        var height = this.element.outerHeight(true);
        var realStop = this._stop - height;

        if (this._parentHeight - this._offset > height && !this._frozen) {
            if (scrollTop >= this._start && scrollTop <= realStop) {
                this.updateOffset(scrollTop - this._start);
            } else {
                if (scrollTop < this._start) {
                    this.updateOffset(0);
                } else if (scrollTop > realStop) {
                    this.updateOffset(this._parentHeight - height - this._offset);
                }
            }
        }
    };

    /**
     * This will set the boundaries the stickied item can move between and it's left position
     * @param [offset] {Number} Manually set the element offset
     */
    Sticky.prototype.setBoundaries = function(offset) {
        this._offset = typeof offset === 'undefined' ? this.element.position().top : offset;
        this._start = this.parent.offset().top + this._offset;
        this._parentHeight = this.parent[0].offsetHeight;
        this._stop = this._start + this._parentHeight - this._offset;
    };

    /**
     * Update Stickied Element's offset thereby moving it up/down the page
     * @param yOffset {Number}
     */
    Sticky.prototype.updateOffset = function(yOffset) {
        if (this._lastPosition !== yOffset) {
            if (this.options.animate) {
                this.element.stop(true, false).delay(this.options.animDelay).animate({
                    'top': yOffset
                }, this.animTime);
            } else {
                this.element.css('top', yOffset);
            }

            this._lastPosition = yOffset;
        }
    };

    /**
     * This will freeze/unfreeze the stickied item
     */
    Sticky.prototype.toggleFreeze = function() {
        this._frozen = !this._frozen;
        this.element.stop(true, false);
        if(!this._frozen){
            this.moveIt();
        }
    };

    $.fn.sticky = function(parentName, options) {
        var method = parentName;
        var ret = false;
        var args = [];

        if (typeof options === 'string') {
            args = [].slice.call(arguments, 0);
        }

        this.each(function() {
            var self = $(this);
            var instance = self.data('stickyInstance');

            if (instance && (options || method)) {
                if (typeof options === 'object') {
                    ret = $.extend(instance.options, options);
                } else if (options === 'options') {
                    ret = instance.options;
                } else if (typeof instance[method] === 'function') {
                    ret = instance[method].apply(instance, args.slice(1));
                } else {
                    console.error('Sticky Element has no option/method named ' + method);
                }
            } else {
                var parent = null;
                if (parent) {
                    parent = self.parent().closest(parent);
                } else {
                    parent = self.parent();
                }

                instance = new Sticky(self, parent, options || {});
                self.data('stickyInstance', instance);
                $.fn.sticky._instances.push(instance);
            }
        });
        return ret || this;
    };

    /**
     * Update the position/offset changed on resize and move, applies to all instances
     */
    var updateAll = function(){
        var len = $.fn.sticky._instances.length;
        for(var i = 0; i < len; i++){
            $.fn.sticky._instances[i].update();
        }
    };

    $.fn.sticky._instances = [];
    $.fn.sticky.updateAll = updateAll;

    $(win).on({
        'resize': function() {
            // Update the boundaries is the browser window is resized
            updateAll();
        },
        'scroll': function() {
            // Move each unfrozen instance on scroll
            var len = $.fn.sticky._instances.length;
            for(var i = 0; i < len; i++){
                var element = $.fn.sticky._instances[i];
                if (!element._frozen) {
                    element.moveIt();
                }
            }
        }
    });

    $(win.document).on({
        'ready': function(){
            // Start an interval to check the heights of sticky elements and update boundaries if necessary
            win.setInterval(function(){
                requestFrame(function(){
                    var len = $.fn.sticky._instances.length;
                    for(var i = 0; i < len; i++){
                        var element = $.fn.sticky._instances[i];
                        if(element._parentHeight !== element.parent[0].offsetHeight){
                            element.update();
                        }
                    }
                });
            },1000);
        }
    })
}(jQuery, window));